import click
import docker
from colorama import Fore, Style

from vantage6.common import info, warning, error
from vantage6.common.docker.addons import (
    check_docker_running,
    remove_container_if_exists,
)
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext


@click.command()
@click_insert_context(InstanceType.ALGORITHM_STORE)
@click.option("--all", "all_servers", flag_value=True, help="Stop all servers")
def cli_algo_store_stop(ctx: AlgorithmStoreContext, all_servers: bool):
    """
    Stop one or all running server(s).
    """
    check_docker_running()
    client = docker.from_env()

    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.ALGORITHM_STORE}"}
    )

    if not running_servers:
        warning("No algorithm stores are currently running.")
        return

    running_server_names = [server.name for server in running_servers]

    if all_servers:
        for container_name in running_server_names:
            _stop_algorithm_store(client, container_name)
        return

    container_name = ctx.docker_container_name
    if container_name not in running_server_names:
        error(f"{Fore.RED}{ctx.name}{Style.RESET_ALL} is not running!")
        return

    _stop_algorithm_store(client, container_name)


def _stop_algorithm_store(client, container_name) -> None:
    """
    Stop the algorithm store server.

    Parameters
    ----------
    client : DockerClient
        The docker client
    container_name : str
        The name of the container to stop
    """
    remove_container_if_exists(client, name=container_name)
    info(f"Stopped the {Fore.GREEN}{container_name}{Style.RESET_ALL} server.")



import docker
import json
import os
import shutil
import base64
import hashlib
import time

from pathlib import Path

from vantage6.common.globals import APPNAME
from vantage6.common import debug, info, error
from vantage6.common.docker.addons import get_container
from vantage6.common.docker.network_manager import NetworkManager
from vantage6.cli.context.server import ServerContext
from vantage6.cli.rabbitmq.definitions import RABBITMQ_DEFINITIONS
from vantage6.cli.globals import RABBIT_TIMEOUT
from vantage6.cli.rabbitmq import split_rabbitmq_uri

DEFAULT_RABBIT_IMAGE = "harbor2.vantage6.ai/infrastructure/rabbitmq"
RABBIT_CONFIG = "rabbitmq.config"
RABBIT_DIR = "rabbitmq"


class RabbitMQManager:
    """
    Manages the RabbitMQ docker container

    Parameters
    ----------
    ctx: ServerContext
        Configuration object
    network_mgr: NetworkManager
        Network manager for network in which server container resides
    image: str
        Docker image to use for RabbitMQ container. By default, the image
        harbor2.vantage6.ai/infrastructure/rabbitmq is used.
    """

    def __init__(
        self, ctx: ServerContext, network_mgr: NetworkManager, image: str = None
    ) -> None:
        self.ctx = ctx
        self.queue_uri = self.ctx.config["rabbitmq"].get("uri")
        rabbit_splitted = split_rabbitmq_uri(self.queue_uri)
        self.rabbit_user = rabbit_splitted["user"]
        self.rabbit_pass = rabbit_splitted["password"]
        self.vhost = rabbit_splitted["vhost"]
        self.port = rabbit_splitted["port"]
        self.host = rabbit_splitted["host"]
        self.definitions_file = Path(self.ctx.data_dir / "definitions.json")
        self.network_mgr = network_mgr

        self.docker = docker.from_env()
        self.image = image if image else DEFAULT_RABBIT_IMAGE

    def start(self) -> None:
        """
        Start a docker container which runs a RabbitMQ queue
        """
        # get volumes which contain rabbitMQ configuration and a location to
        # save rabbitMQ files so that they persist after container is stopped
        volumes = self._get_volumes()

        # expose port 5672 inside the container as port 5672 on the host, and
        # same for 15672 in container to 8080 on host
        # TODO check if these ports are not already used on the host
        ports = {
            f"{self.port}/tcp": self.port,
            # TODO this is for the management tool, do we keep this? Not used
            # at the moment..
            "15672/tcp": 8080,
        }

        # check if a RabbitMQ container is already running
        self.rabbit_container = get_container(docker_client=self.docker, name=self.host)
        if self.rabbit_container:
            info("RabbitMQ is already running! Linking the server to that " "queue")
            if not self.network_mgr.contains(self.rabbit_container):
                self.network_mgr.connect(self.rabbit_container)
            return

        # start rabbitMQ container
        self.rabbit_container = self.docker.containers.run(
            name=self.host,
            image=self.image,
            volumes=volumes,
            ports=ports,
            detach=True,
            restart_policy={"Name": "always"},
            hostname=self.host,
            labels={
                f"{APPNAME}-type": "rabbitmq",
            },
            network=self.network_mgr.network_name,
        )

        # Wait until RabbitMQ is up before continuing with other stuff
        self._wait_for_startup()

    def _wait_for_startup(self) -> None:
        """Wait until RabbitMQ has been initialized"""
        interval = 10
        attempts = int((RABBIT_TIMEOUT + interval) / interval)
        is_running = False
        for _ in range(attempts):
            if self.is_running():
                is_running = True
                break
            debug(f"RabbitMQ is not yet running. Retrying in {interval}s...")
            time.sleep(interval)
        if is_running:
            info("RabbitMQ was started successfully!")
        else:
            error("Could not start RabbitMQ! Exiting...")
            exit(1)

    def is_running(self) -> bool:
        """
        Returns
        -------
        bool
            Whether the container has fully initialized RabbitMQ or not
        """
        response = self.rabbit_container.exec_run(
            cmd="rabbitmqctl status --formatter json"
        )
        return response.exit_code == 0

    def _get_volumes(self) -> dict:
        """
        Prepare the volumes for the RabbitMQ container. The RabbitMQ should
        set up the right vhost and users to allow the server to communicate
        with RabbitMQ as configured.

        Returns
        -------
        dict
            Dictionary with the volumes to mount in the RabbitMQ container
        """
        # default RabbitMQ configuration: replace the user/password with the
        # credentials from the configuraiton
        rabbit_definitions = self._get_rabbitmq_definitions()

        # write the RabbitMQ definition to file(s)
        with open(self.definitions_file, "w") as f:
            json.dump(rabbit_definitions, f, indent=2)

        # write RabbitMQ config to file
        rabbit_conf = Path(__file__).parent.resolve() / RABBIT_CONFIG
        shutil.copyfile(rabbit_conf, self.ctx.data_dir / RABBIT_CONFIG)

        # check if a directory for persistent RabbitMQ storage exists,
        # otherwise create it
        rabbit_data_dir = self.ctx.data_dir / RABBIT_DIR
        if not rabbit_data_dir.exists():
            rabbit_data_dir.mkdir(parents=True, exist_ok=True)

        return {
            self.definitions_file: {
                "bind": "/etc/rabbitmq/definitions.json",
                "mode": "ro",
            },
            self.ctx.data_dir
            / RABBIT_CONFIG: {"bind": "/etc/rabbitmq/rabbitmq.config", "mode": "ro"},
            rabbit_data_dir: {"bind": "/var/lib/rabbitmq", "mode": "rw"},
        }

    def _get_rabbitmq_definitions(self) -> dict:
        """
        Get startup definitions (users/vhosts etc) for RabbitMQ container

        Returns
        -------
        Dict:
            dictionary with all users/vhosts etc that must be generated on
            startup of RabbitMQ
        """
        rabbit_definitions = RABBITMQ_DEFINITIONS
        rabbit_definitions["users"][0]["name"] = self.rabbit_user
        rabbit_definitions["permissions"][0]["user"] = self.rabbit_user
        rabbit_definitions["users"][0]["password_hash"] = self._get_hashed_pw(
            self.rabbit_pass
        )
        rabbit_definitions["vhosts"][0]["name"] = self.vhost
        rabbit_definitions["permissions"][0]["vhost"] = self.vhost
        return rabbit_definitions

    def _get_hashed_pw(self, pw: str) -> str:
        """
        Hash a user-defined password for RabbitMQ

        Parameters
        ----------
        pw: str
            The password for RabbitMQ

        Returns
        -------
        str
            Hashed password for RabbitMQ
        """

        # Generate a random 32 bit salt:
        salt = os.urandom(4)

        # Concatenate that with the UTF-8 representation of the password
        tmp0 = salt + pw.encode("utf-8")

        # Take the SHA256 hash and get the bytes back
        tmp1 = hashlib.sha256(tmp0).digest()

        # Concatenate the salt again:
        salted_hash = salt + tmp1

        # convert to base64 encoding:
        pass_hash = base64.b64encode(salted_hash)
        return pass_hash.decode("utf-8")



import click
import docker

from colorama import Fore, Style

from vantage6.common import error
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import attach_logs
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext


@click.command()
@click_insert_context(InstanceType.ALGORITHM_STORE)
def cli_algo_store_attach(ctx: AlgorithmStoreContext) -> None:
    """
    Show the server logs in the current console.
    """
    check_docker_running()
    client = docker.from_env()

    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.ALGORITHM_STORE}"}
    )
    running_server_names = [container.name for container in running_servers]

    if ctx.docker_container_name in running_server_names:
        container = client.containers.get(ctx.docker_container_name)
        attach_logs(container, InstanceType.ALGORITHM_STORE)
    else:
        error(f"{Fore.RED}{ctx.name}{Style.RESET_ALL} is not running!")



import sys
import click

from vantage6.client import UserClient
from vantage6.cli.utils import error
from vantage6.cli.test.common.diagnostic_runner import DiagnosticRunner


@click.command()
@click.option("--host", type=str, default="http://localhost", help="URL of the server")
@click.option("--port", type=int, default=5000, help="Port of the server")
@click.option("--api-path", type=str, default="/api", help="API path of the server")
@click.option(
    "--username",
    type=str,
    default="root",
    help="Username of vantage6 user account to create the task with",
)
@click.option(
    "--password",
    type=str,
    default="root",
    help="Password of vantage6 user account to create the task with",
)
@click.option(
    "--collaboration",
    type=int,
    default=1,
    help="ID of the collaboration to create the task in",
)
@click.option(
    "-o",
    "--organizations",
    type=int,
    default=[],
    multiple=True,
    help="ID(s) of the organization(s) to create the task for",
)
@click.option(
    "--all-nodes",
    is_flag=True,
    help="Run the diagnostic test on all nodes in the collaboration",
)
@click.option(
    "--online-only",
    is_flag=True,
    help="Run the diagnostic test on only nodes that are online",
)
@click.option("--no-vpn", is_flag=True, help="Don't execute VPN tests")
@click.option(
    "--private-key",
    type=str,
    default=None,
    help="Path to the private key for end-to-end encryption",
)
def cli_test_features(
    host: str,
    port: int,
    api_path: str,
    username: str,
    password: str,
    collaboration: int,
    organizations: list[int] | None,
    all_nodes: bool,
    online_only: bool,
    no_vpn: bool,
    private_key: str | None,
) -> list[dict]:
    """
    Run diagnostic checks on an existing vantage6 network.

    This command will create a task in the requested collaboration that will
    test the functionality of vantage6, and will report back the results.
    """
    if all_nodes and organizations:
        error("Cannot use --all-nodes and --organization at the same time.")
        sys.exit(1)

    if all_nodes or not organizations:
        organizations = None

    client = UserClient(host=host, port=port, path=api_path, log_level="critical")
    client.authenticate(username=username, password=password)
    client.setup_encryption(private_key)
    diagnose = DiagnosticRunner(client, collaboration, organizations, online_only)
    res = diagnose(base=True, vpn=not no_vpn)
    return res



import subprocess

import click
import docker
from colorama import Fore, Style

from vantage6.common import info, error, debug as debug_msg
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.cli.context.server import ServerContext
from vantage6.cli.common.decorator import click_insert_context


@click.command()
@click_insert_context(type_="server")
def cli_server_shell(ctx: ServerContext) -> None:
    """
    Run an iPython shell within a running server. This can be used to modify
    the database.

    NOTE: using the shell is no longer recommended as there is no validation on
    the changes that you make. It is better to use the Python client or a
    graphical user interface instead.
    """
    # will print an error if not
    check_docker_running()

    docker_client = docker.from_env()

    running_servers = docker_client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.SERVER}"}
    )

    if ctx.docker_container_name not in [s.name for s in running_servers]:
        error(f"Server {Fore.RED}{ctx.name}{Style.RESET_ALL} is not running?")
        return

    try:
        subprocess.run(
            [
                "docker",
                "exec",
                "-it",
                ctx.docker_container_name,
                "vserver-local",
                "shell",
                "-c",
                "/mnt/config.yaml",
            ]
        )
    except Exception as e:
        info("Failed to start subprocess...")
        debug_msg(e)



from pathlib import Path
import csv
import yaml
import click
from jinja2 import Environment, FileSystemLoader
from colorama import Fore, Style

from vantage6.common.globals import APPNAME, InstanceType
from vantage6.common import info, error, generate_apikey

from vantage6.cli.globals import PACKAGE_FOLDER
from vantage6.cli.context.server import ServerContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.server.common import get_server_context
from vantage6.cli.server.import_ import cli_server_import
from vantage6.cli.utils import prompt_config_name


def create_dummy_data(node_name: str, dev_folder: Path) -> Path:
    """Synthesize csv dataset.

    Parameters
    ----------
    node_name : str
        Name of node to be used as part of dataset.
    dev_folder : Path
        Path to the dev folder.

    Returns
    -------
    Path
        Directory the data is saved in.
    """
    header = ["name", "mask", "weapon", "age"]
    data = [
        ["Raphael", "red", "sai", 44],
        ["Donatello", "purple", "bo staff", 60],
    ]

    data_file = dev_folder / f"df_{node_name}.csv"
    with open(data_file, "w", encoding="UTF8", newline="") as f:
        writer = csv.writer(f)

        # write the header
        writer.writerow(header)

        # write the data
        for row in data:
            writer.writerow(row)

        f.close()

    info(
        f"Spawned dataset for {Fore.GREEN}{node_name}{Style.RESET_ALL}, "
        f"writing to {Fore.GREEN}{data_file}{Style.RESET_ALL}"
    )
    return data_file


def create_node_config_file(
    server_url: str, port: int, config: dict, server_name: str
) -> None:
    """Create a node configuration file (YAML).

    Creates a node configuration for a simulated organization. Organization ID
    is used for generating both the organization name and node_name as each
    organization only houses one node.

    Parameters
    ----------
    server_url : str
        Url of the dummy server.
    port : int
        Port of the dummy server.
    config : dict
        Configuration dictionary containing org_id, api_key, node name and
        additional user_defined_config.
    server_name : str
        Configuration name of the dummy server.
    """
    environment = Environment(
        loader=FileSystemLoader(PACKAGE_FOLDER / APPNAME / "cli" / "template"),
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=True,
    )
    template = environment.get_template("node_config.j2")

    # TODO: make this name specific to the server it connects
    node_name = config["node_name"]
    folders = NodeContext.instance_folders("node", node_name, False)
    path_to_dev_dir = Path(folders["dev"] / server_name)
    path_to_dev_dir.mkdir(parents=True, exist_ok=True)
    dummy_datafile = create_dummy_data(node_name, path_to_dev_dir)

    path_to_data_dir = Path(folders["data"])
    path_to_data_dir.mkdir(parents=True, exist_ok=True)
    full_path = Path(folders["config"] / f"{node_name}.yaml")

    if full_path.exists():
        error(f"Node configuration file already exists: {full_path}")
        exit(1)

    node_config = template.render(
        {
            "api_key": config["api_key"],
            "databases": {"default": dummy_datafile},
            "logging": {"file": f"{node_name}.log"},
            "port": port,
            "server_url": server_url,
            "task_dir": str(path_to_data_dir),
            "user_provided_config": config["user_defined_config"],
        }
    )

    try:
        with open(full_path, "x") as f:
            f.write(node_config)
    except Exception as e:
        error(f"Could not write node configuration file: {e}")
        exit(1)

    info(
        f"Spawned node for organization {Fore.GREEN}{config['org_id']}"
        f"{Style.RESET_ALL}"
    )


def _read_extra_config_file(extra_config_file: Path | None) -> str:
    """Reads extra configuration file.

    Parameters
    ----------
    extra_config_file : Path | None
        Path to file with additional configuration.

    Returns
    -------
    str
        Extra configuration file content
    """
    if extra_config_file:
        # read the YAML file as string, so it can be appended to the
        # configuration easily
        with open(extra_config_file, "r", encoding="utf-8") as f:
            return f.read()
    return ""


def generate_node_configs(
    num_nodes: int,
    server_url: str,
    port: int,
    server_name: str,
    extra_node_config: Path | None,
) -> list[dict]:
    """Generates ``num_nodes`` node configuration files.

    Parameters
    ----------
    num_nodes : int
        Integer to determine how many configurations to create.
    server_url : str
        Url of the dummy server.
    port : int
        Port of the dummy server.
    server_name : str
        Configuration name of the dummy server.
    extra_node_config : Path | None
        Path to file with additional node configuration.

    Returns
    -------
    list[dict]
        List of dictionaries containing node configurations.
    """
    configs = []
    extra_config = _read_extra_config_file(extra_node_config)
    for i in range(num_nodes):
        config = {
            "org_id": i + 1,
            "api_key": generate_apikey(),
            "node_name": f"{server_name}_node_{i + 1}",
            "user_defined_config": extra_config,
        }
        create_node_config_file(server_url, port, config, server_name)
        configs.append(config)

    return configs


def create_vserver_import_config(node_configs: list[dict], server_name: str) -> Path:
    """Create server configuration import file (YAML).

    Utilized by the ``v6 server import`` command.

    Parameters
    ----------
    node_configs : list[dict]
        List of dictionaries containing the node configurations, returned from
        ``generate_node_configs()``.
    server_name : str
        Server name.

    Returns
    -------
    Path
        Path object where the server import configuration is stored.
    """
    environment = Environment(
        loader=FileSystemLoader(PACKAGE_FOLDER / APPNAME / "cli" / "template"),
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=True,
    )
    template = environment.get_template("server_import_config.j2")

    organizations = []
    collaboration = {"name": "demo", "participants": []}
    for config in node_configs:
        org_id = config["org_id"]
        org_data = {"name": f"org_{org_id}"}

        organizations.append(org_data)
        collaboration["participants"].append(
            {"name": f"org_{org_id}", "api_key": config["api_key"]}
        )
    organizations[0]["make_admin"] = True
    info(
        f"Organization {Fore.GREEN}{node_configs[0]['org_id']}"
        f"{Style.RESET_ALL} is the admin"
    )

    server_import_config = template.render(
        organizations=organizations, collaboration=collaboration
    )
    folders = ServerContext.instance_folders(InstanceType.SERVER, server_name, False)

    demo_dir = Path(folders["dev"])
    demo_dir.mkdir(parents=True, exist_ok=True)
    full_path = demo_dir / f"{server_name}.yaml"
    if full_path.exists():
        error(f"Server configuration file already exists: {full_path}")
        exit(1)

    try:
        with open(full_path, "x") as f:
            f.write(server_import_config)
            info(
                "Server import configuration ready, writing to "
                f"{Fore.GREEN}{full_path}{Style.RESET_ALL}"
            )
    except Exception as e:
        error(f"Could not write server import configuration file: {e}")
        exit(1)

    return full_path


def create_vserver_config(server_name: str, port: int, extra_config_file: Path) -> Path:
    """Creates server configuration file (YAML).

    Parameters
    ----------
    server_name : str
        Server name.
    port : int
        Server port.
    extra_config_file : Path
        Path to file with additional server configuration.

    Returns
    -------
    Path
        Path object where server configuration is stored.
    """
    environment = Environment(
        loader=FileSystemLoader(PACKAGE_FOLDER / APPNAME / "cli" / "template"),
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=True,
    )

    extra_config = _read_extra_config_file(extra_config_file)

    template = environment.get_template("server_config.j2")
    server_config = template.render(
        port=port, jwt_secret_key=generate_apikey(), user_provided_config=extra_config
    )
    folders = ServerContext.instance_folders(
        instance_type="server", instance_name=server_name, system_folders=True
    )

    config_dir = Path(folders["config"] / server_name)
    config_dir.mkdir(parents=True, exist_ok=True)
    full_path = folders["config"] / f"{server_name}.yaml"
    if full_path.exists():
        error(f"Server configuration file already exists: {full_path}")
        exit(1)

    try:
        with open(full_path, "x") as f:
            f.write(server_config)
            info(
                "Server configuration read, writing to "
                f"{Fore.GREEN}{full_path}{Style.RESET_ALL}"
            )
    except Exception as e:
        error(f"Could not write server configuration file: {e}")
        exit(1)

    return full_path


def demo_network(
    num_nodes: int,
    server_url: str,
    server_port: int,
    server_name: str,
    extra_server_config: Path,
    extra_node_config: Path,
) -> tuple[list[dict], Path, Path]:
    """Generates the demo network.

    Parameters
    ----------
    num_nodes : int
        Integer to determine how many configurations to create.
    server_url : str
        Url of the dummy server.
    server_port : int
        Port of the dummy server.
    server_name : str
        Server name.
    extra_server_config : Path
        Path to file with additional server configuration.
    extra_node_config : Path
        Path to file with additional node configuration.

    Returns
    -------
    tuple[list[dict], Path, Path]
        Tuple containing node, server import and server configurations.
    """
    node_configs = generate_node_configs(
        num_nodes, server_url, server_port, server_name, extra_node_config
    )
    server_import_config = create_vserver_import_config(node_configs, server_name)
    server_config = create_vserver_config(server_name, server_port, extra_server_config)
    return (node_configs, server_import_config, server_config)


@click.command()
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your development setup"
)
@click.option(
    "--num-nodes",
    type=int,
    default=3,
    help="Generate this number of nodes in the development network",
)
@click.option(
    "--server-url",
    type=str,
    default="http://host.docker.internal",
    help="Server URL to point to. If you are using Docker Desktop, "
    "the default http://host.docker.internal should not be changed.",
)
@click.option(
    "-p",
    "--server-port",
    type=int,
    default=5000,
    help="Port to run the server on. Default is 5000.",
)
@click.option(
    "-i",
    "--image",
    type=str,
    default=None,
    help="Server docker image to use when setting up resources for "
    "the development server",
)
@click.option(
    "--extra-server-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional server "
    "configuration. This will be appended to the server "
    "configuration file",
)
@click.option(
    "--extra-node-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional node configuration. This will be"
    " appended to each of the node configuration files",
)
@click.pass_context
def create_demo_network(
    click_ctx: click.Context,
    name: str,
    num_nodes: int,
    server_url: str,
    server_port: int,
    image: str = None,
    extra_server_config: Path = None,
    extra_node_config: Path = None,
) -> dict:
    """Creates a demo network.

    Creates server instance as well as its import configuration file. Server
    name is set to 'dev_default_server'. Generates `n` node configurations, but
    by default this is set to 3. Then runs a Batch import of
    organizations/collaborations/users and tasks.
    """
    server_name = prompt_config_name(name)
    if not ServerContext.config_exists(server_name):
        demo = demo_network(
            num_nodes,
            server_url,
            server_port,
            server_name,
            extra_server_config,
            extra_node_config,
        )
        info(
            f"Created {Fore.GREEN}{len(demo[0])}{Style.RESET_ALL} node "
            f"configuration(s), attaching them to {Fore.GREEN}{server_name}"
            f"{Style.RESET_ALL}."
        )
    else:
        error(f"Configuration {Fore.RED}{server_name}{Style.RESET_ALL} already exists!")
        exit(1)
    (node_config, server_import_config, server_config) = demo
    ctx = get_server_context(server_name, True, ServerContext)
    click_ctx.invoke(
        cli_server_import,
        ctx=ctx,
        file=server_import_config,
        drop_all=False,
        image=image,
        mount_src="",
        keep=False,
        wait=True,
    )
    info(
        "Development network was set up successfully! You can now start the "
        f"server and nodes with {Fore.GREEN}v6 server start-demo-network"
        f"{Style.RESET_ALL}"
    )
    # find user credentials to print. Read from server import file
    with open(server_import_config, "r") as f:
        server_import_config = yaml.safe_load(f)

    try:
        user = server_import_config["organizations"][0]["users"][0]
        username = user["username"]
        password = user["password"]
        info(
            "You can login with the following credentials:\n"
            f"Username: {username}\n"
            f"Password: {password}\n"
        )
    except KeyError:
        # No user found, skip printing credentials
        pass

    return {
        "node_configs": node_config,
        "server_import_config": server_import_config,
        "server_config": server_config,
    }



import questionary as q

from pathlib import Path

from vantage6.common import generate_apikey
from vantage6.common.globals import DATABASE_TYPES, InstanceType
from vantage6.common.client.node_client import NodeClient
from vantage6.common.context import AppContext
from vantage6.common import error, warning, info
from vantage6.cli.context import select_context_class
from vantage6.cli.configuration_manager import (
    NodeConfigurationManager,
    ServerConfigurationManager,
)
from vantage6.cli.globals import AlgoStoreGlobals, ServerGlobals


def node_configuration_questionaire(dirs: dict, instance_name: str) -> dict:
    """
    Questionary to generate a config file for the node instance.

    Parameters
    ----------
    dirs : dict
        Dictionary with the directories of the node instance.
    instance_name : str
        Name of the node instance.

    Returns
    -------
    dict
        Dictionary with the new node configuration
    """
    config = q.prompt(
        [
            {"type": "text", "name": "api_key", "message": "Enter given api-key:"},
            {
                "type": "text",
                "name": "server_url",
                "message": "The base-URL of the server:",
                "default": "http://localhost",
            },
            {
                "type": "text",
                "name": "port",
                "message": "Enter port to which the server listens:",
                "default": "5000",
            },
            {
                "type": "text",
                "name": "api_path",
                "message": "Path of the api:",
                "default": "/api",
            },
            {
                "type": "text",
                "name": "task_dir",
                "message": "Task directory path:",
                "default": str(dirs["data"]),
            },
        ]
    )

    config["databases"] = list()
    while q.confirm("Do you want to add a database?").ask():
        db_label = q.prompt(
            [
                {
                    "type": "text",
                    "name": "label",
                    "message": "Enter unique label for the database:",
                    "default": "default",
                }
            ]
        )
        db_path = q.prompt(
            [{"type": "text", "name": "uri", "message": "Database URI:"}]
        )
        db_type = q.select("Database type:", choices=DATABASE_TYPES).ask()

        config["databases"].append(
            {"label": db_label.get("label"), "uri": db_path.get("uri"), "type": db_type}
        )

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
    ).ask()

    is_add_vpn = q.confirm(
        "Do you want to connect to a VPN server?", default=False
    ).ask()
    if is_add_vpn:
        config["vpn_subnet"] = q.text(
            message="Subnet of the VPN server you want to connect to:",
            default="10.76.0.0/16",
        ).ask()

    is_policies = q.confirm(
        "Do you want to add limit the algorithms allowed to run on your node? This "
        "should always be done for production scenarios.",
        default=True,
    ).ask()
    if is_policies:
        allowed_algorithms = []
        info("Below you can add algorithms that are allowed to run on your node.")
        info(
            "You can use regular expressions to match multiple algorithms, or you can "
            "use strings to provide one algorithm at a time."
        )
        info("Examples:")
        # pylint: disable=W1401
        # flake8: noqa: W605
        info("^harbor2\.vantage6\.ai/demo/average$    Allow the demo average algorithm")
        info(
            "^harbor2\.vantage6\.ai/algorithms/.*   Allow all algorithms from "
            "harbor2.vantage6.ai/algorithms"
        )
        info(
            "^harbor2\.vantage6\.ai/demo/average:@sha256:82becede...$    Allow a "
            "specific hash of average algorithm"
        )
        while True:
            algo = q.text(message="Enter your algorithm expression:").ask()
            allowed_algorithms.append(algo)
            if not q.confirm(
                "Do you want to add another algorithm expression?", default=True
            ).ask():
                break
        config["policies"] = {"allowed_algorithms": allowed_algorithms}

    config["logging"] = {
        "level": res,
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "requests", "level": "warning"},
            {"name": "engineio.client", "level": "warning"},
            {"name": "docker.utils.config", "level": "warning"},
            {"name": "docker.auth", "level": "warning"},
        ],
    }

    # Check if we can login to the server to retrieve collaboration settings
    client = NodeClient(config["server_url"], config["port"], config["api_path"])
    try:
        client.authenticate(config["api_key"])
    except Exception as e:
        error(f"Could not authenticate with server: {e}")
        error("Please check (1) your API key and (2) if your server is online")
        warning(
            "If you continue, you should provide your collaboration "
            "settings manually."
        )
        if q.confirm("Do you want to abort?", default=True).ask():
            exit(0)

    if client.whoami is not None:
        encryption = client.is_encrypted_collaboration()
        # TODO when we build collaboration policies, update this to provide
        # the node admin with a list of all policies, and whether or not
        # to accept them
        q.confirm(
            f"Encryption is {'enabled' if encryption else 'disabled'}"
            f" for this collaboration. Accept?",
            default=True,
        ).ask()
    else:
        encryption = q.confirm("Enable encryption?", default=True).ask()

    private_key = "" if not encryption else q.text("Path to private key file:").ask()

    config["encryption"] = {
        "enabled": encryption is True or encryption == "true",
        "private_key": private_key,
    }

    return config


def _get_common_server_config(
    instance_type: InstanceType, instance_name: str, include_api_path: bool = True
) -> dict:
    """
    Part of the questionaire that is common to all server types (vantage6
    server and algorithm store server).

    Parameters
    ----------
    instance_type : InstanceType
        Type of server instance.
    instance_name : str
        Name of the server instance.
    include_api_path : bool
        Whether to include the api path in the questionaire.

    Returns
    -------
    dict
        Dictionary with new (partial) server configuration
    """
    config = q.prompt(
        [
            {
                "type": "text",
                "name": "description",
                "message": "Enter a human-readable description:",
            },
            {"type": "text", "name": "ip", "message": "ip:", "default": "0.0.0.0"},
            {
                "type": "text",
                "name": "port",
                "message": "Enter port to which the server listens:",
                "default": (
                    # Note that .value is required in YAML to get proper str value
                    ServerGlobals.PORT.value
                    if instance_type == InstanceType.SERVER
                    else AlgoStoreGlobals.PORT.value
                ),
            },
        ]
    )

    # TODO v5+ remove api_path. It complicates configuration
    if include_api_path:
        config.update(
            q.prompt(
                [
                    {
                        "type": "text",
                        "name": "api_path",
                        "message": "Path of the api:",
                        "default": "/api",
                    }
                ]
            )
        )

    config.update(
        q.prompt(
            [
                {
                    "type": "text",
                    "name": "uri",
                    "message": "Database URI:",
                    "default": "sqlite:///default.sqlite",
                },
                {
                    "type": "select",
                    "name": "allow_drop_all",
                    "message": "Allowed to drop all tables: ",
                    "choices": ["True", "False"],
                },
            ]
        )
    )

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
    ).ask()

    config["logging"] = {
        "level": res,
        "file": f"{instance_name}.log",
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "socketIO-client", "level": "warning"},
            {"name": "socketio.server", "level": "warning"},
            {"name": "engineio.server", "level": "warning"},
            {"name": "sqlalchemy.engine", "level": "warning"},
            {"name": "requests_oauthlib.oauth2_session", "level": "warning"},
        ],
    }

    return config


def server_configuration_questionaire(instance_name: str) -> dict:
    """
    Questionary to generate a config file for the server instance.

    Parameters
    ----------
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with the new server configuration
    """

    config = _get_common_server_config(
        InstanceType.SERVER, instance_name, include_api_path=True
    )

    constant_jwt_secret = q.confirm("Do you want a constant JWT secret?").ask()
    if constant_jwt_secret:
        config["jwt_secret_key"] = generate_apikey()

    is_mfa = q.confirm("Do you want to enforce two-factor authentication?").ask()
    if is_mfa:
        config["two_factor_auth"] = is_mfa

    current_server_url = f"http://localhost:{config['port']}{config['api_path']}"
    config["server_url"] = q.text(
        "What is the server url exposed to the users? If you are running a"
        " development server running locally, this is the same as the "
        "server url. If you are running a production server, this is the "
        "url that users will connect to.",
        default=current_server_url,
    ).ask()

    is_add_vpn = q.confirm("Do you want to add a VPN server?", default=False).ask()
    if is_add_vpn:
        vpn_config = q.prompt(
            [
                {
                    "type": "text",
                    "name": "url",
                    "message": "VPN server URL:",
                },
                {
                    "type": "text",
                    "name": "portal_username",
                    "message": "VPN portal username:",
                },
                {
                    "type": "password",
                    "name": "portal_userpass",
                    "message": "VPN portal password:",
                },
                {
                    "type": "text",
                    "name": "client_id",
                    "message": "VPN client username:",
                },
                {
                    "type": "password",
                    "name": "client_secret",
                    "message": "VPN client password:",
                },
                {
                    "type": "text",
                    "name": "redirect_url",
                    "message": "Redirect url (should be local address of server)",
                    "default": "http://localhost",
                },
            ]
        )
        config["vpn_server"] = vpn_config

    is_add_rabbitmq = q.confirm("Do you want to add a RabbitMQ message queue?").ask()
    if is_add_rabbitmq:
        rabbit_uri = q.text(message="Enter the URI for your RabbitMQ:").ask()
        run_rabbit_locally = q.confirm(
            "Do you want to run RabbitMQ locally? (Use only for testing)"
        ).ask()
        config["rabbitmq"] = {
            "uri": rabbit_uri,
            "start_with_server": run_rabbit_locally,
        }

    # add algorithm stores to this server
    is_add_community_store = q.confirm(
        "Do you want to make the algorithms from the community algorithm store "
        "available to your users?"
    ).ask()
    algorithm_stores = []
    if is_add_community_store:
        algorithm_stores.append(
            {"name": "Community store", "url": "https://store.cotopaxi.vantage6.ai"}
        )
    add_more_stores = q.confirm(
        "Do you want to add more algorithm stores?", default=False
    ).ask()
    while add_more_stores:
        store_name = q.text(message="Enter the name of the store:").ask()
        store_url = q.text(message="Enter the URL of the store:").ask()
        algorithm_stores.append({"name": store_name, "url": store_url})
        add_more_stores = q.confirm(
            "Do you want to add more algorithm stores?", default=False
        ).ask()
    config["algorithm_stores"] = algorithm_stores

    return config


def algo_store_configuration_questionaire(instance_name: str) -> dict:
    """
    Questionary to generate a config file for the algorithm store server
    instance.

    Parameters
    ----------
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with the new server configuration
    """
    config = _get_common_server_config(
        InstanceType.ALGORITHM_STORE, instance_name, include_api_path=False
    )

    default_v6_server_uri = "http://localhost:5000/api"
    default_root_username = "root"

    v6_server_uri = q.text(
        "What is the Vantage6 server linked to the algorithm store? "
        "Provide the link to the server endpoint.",
        default=default_v6_server_uri,
    ).ask()

    root_username = q.text(
        "What is the username of the root user?",
        default=default_root_username,
    ).ask()

    config["root_user"] = {
        "v6_server_uri": v6_server_uri,
        "username": root_username,
    }

    # ask about openness of the algorithm store
    is_open = q.confirm(
        "Do you want to open the algorithm store to the public? This will allow anyone "
        "to view the algorithms, but they cannot modify them.",
        default=False,
    ).ask()
    if is_open:
        config["policies"] = {"algorithms_open": True}
    else:
        is_open_to_whitelist = q.confirm(
            "Do you want to allow all users of whitelisted vantage6 servers to access "
            "the algorithms in the store? If not allowing this, you will have to assign"
            " the appropriate permissions to each user individually.",
            default=True,
        ).ask()
        config["policies"] = {
            "algorithms_open": False,
            "algorithms_open_to_whitelisted": is_open_to_whitelist,
        }

    return config


def configuration_wizard(
    type_: InstanceType, instance_name: str, system_folders: bool
) -> Path:
    """
    Create a configuration file for a node or server instance.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    instance_name : str
        Name of the instance
    system_folders : bool
        Whether to use the system folders or not

    Returns
    -------
    Path
        Path to the configuration file
    """
    # for defaults and where to save the config
    dirs = AppContext.instance_folders(type_, instance_name, system_folders)

    # invoke questionaire to create configuration file
    if type_ == InstanceType.NODE:
        conf_manager = NodeConfigurationManager
        config = node_configuration_questionaire(dirs, instance_name)
    elif type_ == InstanceType.SERVER:
        conf_manager = ServerConfigurationManager
        config = server_configuration_questionaire(instance_name)
    else:
        conf_manager = ServerConfigurationManager
        config = algo_store_configuration_questionaire(instance_name)

    # in the case of an environment we need to add it to the current
    # configuration. In the case of application we can simply overwrite this
    # key (although there might be environments present)
    config_file = Path(dirs.get("config")) / (instance_name + ".yaml")

    if Path(config_file).exists():
        config_manager = conf_manager.from_file(config_file)
    else:
        config_manager = conf_manager(instance_name)

    config_manager.put(config)
    config_manager.save(config_file)

    return config_file


def select_configuration_questionaire(type_: InstanceType, system_folders: bool) -> str:
    """
    Ask which configuration the user wants to use. It shows only configurations
    that are in the default folder.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    system_folders : bool
        Whether to use the system folders or not

    Returns
    -------
    str
        Name of the configuration
    """
    context = select_context_class(type_)
    configs, _ = context.available_configurations(system_folders)

    # each collection (file) can contain multiple configs. (e.g. test,
    # dev)
    choices = []
    for config_collection in configs:
        choices.append(
            q.Choice(title=f"{config_collection.name:25}", value=config_collection.name)
        )

    if not choices:
        raise Exception("No configurations could be found!")

    # pop the question
    return q.select("Select the configuration you want to use:", choices=choices).ask()



import json

from typing import Any
from rich.console import Console
from rich.table import Table

from vantage6.client import UserClient
from vantage6.common import info, debug
from vantage6.cli.globals import DIAGNOSTICS_IMAGE


class DiagnosticRunner:
    """
    Class to run the diagnostic algorithm on a vantage6 network.

    This class will create a task in the requested collaboration that will test
    the functionality of vantage6, and will report back the results.

    Parameters
    ----------
    client : UserClient
        The client to use for communication with the server.
    collaboration_id : int
        The ID of the collaboration to run the diagnostics in.
    organizations : list[int] | None
        The ID(s) of the organization(s) to run the diagnostics for. If None
        (default), run the diagnostics for all organizations in the
        collaboration.
    online_only : bool
        Whether to run the diagnostics only on nodes that are online. By
        default False
    """

    def __init__(
        self,
        client: UserClient,
        collaboration_id: int,
        organizations: list[int] | None = None,
        online_only: bool = False,
    ) -> None:
        self.client = client
        self.collaboration_id = collaboration_id

        if not organizations:
            # run on all organizations in the collaboration
            # TODO use pagination properly, instead of just getting the first
            # 1000 organizations (which very likely is enough though)
            orgs = self.client.organization.list(
                collaboration=self.collaboration_id, per_page=1000
            )
            self.organization_ids = [org["id"] for org in orgs["data"]]
        else:
            self.organization_ids = organizations

        if online_only:
            nodes = self.client.node.list(
                collaboration=self.collaboration_id, is_online=True
            )
            debug(nodes)
            online_orgs = [node["organization"]["id"] for node in nodes["data"]]
            self.organization_ids = list(
                set(self.organization_ids).intersection(online_orgs)
            )

        info(f"Running diagnostics to {len(self.organization_ids)} " "organization(s)")
        info(f"  organizations: {self.organization_ids}")
        info(f"  collaboration: {self.collaboration_id}")

    def __call__(self, base: bool = True, vpn: bool = True) -> Any:
        """
        Run the diagnostics.

        Parameters
        ----------
        base: bool
            Whether to run the base features of the diagnostic algorithm. By
            default True.
        vpn: bool
            Whether to run the VPN features of the diagnostic algorithm. By
            default True.
        """
        base_results = self.base_features() if base else []
        vpn_results = self.vpn_features() if vpn else []
        return base_results + vpn_results

    def base_features(self) -> list[dict]:
        """
        Create a task to run the base features of the diagnostic algorithm.

        Returns
        -------
        list[dict]
            The results of the diagnostic algorithm.
        """
        task = self.client.task.create(
            collaboration=self.collaboration_id,
            name="test",
            description="Basic Diagnostic test",
            image=DIAGNOSTICS_IMAGE,
            input_={
                "method": "base_features",
            },
            organizations=self.organization_ids,
            databases=[{"label": "default"}],
        )
        debug(task)

        return self._wait_and_display(task.get("id"))

    def vpn_features(self) -> list[dict]:
        """
        Create a task to run the VPN features of the diagnostic algorithm.

        Returns
        -------
        list[dict]
            The results of the diagnostic algorithm.
        """
        self.client.node.list(collaboration=self.collaboration_id)

        task = self.client.task.create(
            collaboration=self.collaboration_id,
            name="test",
            description="VPN Diagnostic test",
            image=DIAGNOSTICS_IMAGE,
            input_={
                "method": "vpn_features",
                "kwargs": {"other_nodes": self.organization_ids},
            },
            organizations=self.organization_ids,
        )

        return self._wait_and_display(task.get("id"))

    def _wait_and_display(self, task_id: int) -> list[dict]:
        """
        Wait for the task to finish and then display the results.

        Parameters
        ----------
        task_id : int
            The ID of the task to wait for.

        Returns
        -------
        list[dict]
            The results of the diagnostic algorithm.
        """
        # TODO should we have the option to combine these in one request? Seems
        # like it would be more efficient
        # TODO ensure that we get all pages of results
        results = self.client.wait_for_results(task_id=task_id)["data"]
        runs = self.client.run.from_task(task_id=task_id)["data"]
        print("\n")
        for res in results:
            matched_run = [run for run in runs if run["id"] == res["run"]["id"]][0]
            self.display_diagnostic_results(res, matched_run["organization"]["id"])
            print()
        return results

    def display_diagnostic_results(self, result: dict, org_id: int) -> None:
        """
        Print the results of the diagnostic algorithm.

        Parameters
        ----------
        result : list[dict]
            The result of the diagnostic algorithm.
        org_id : int
            The ID of the organization that the diagnostic algorithm was run on
        """
        res = json.loads(result["result"])
        t_ = Table(title=f"Basic Diagnostics Summary (organization {org_id})")
        t_.add_column("name")
        t_.add_column("success")
        e_ = Table(title=f"Basic Diagnostics Errors (organization {org_id})")
        e_.add_column("name")
        e_.add_column("exception")
        e_.add_column("traceback")
        e_.add_column("payload")
        errors = False
        for diag in res:
            if diag["success"]:
                success = ":heavy_check_mark: [green]success[/green]"
            else:
                success = ":x: [red]failed[/red]"
                e_.add_row(
                    diag["name"], diag["exception"], diag["traceback"], diag["payload"]
                )
                errors = True
            t_.add_row(diag["name"], success)

        console = Console()
        console.print(t_)
        if errors:
            console.print(e_)



"""RabbitMQ utilities."""


def split_rabbitmq_uri(rabbit_uri: str) -> dict:
    """
    Get details (user, pass, host, vhost, port) from a RabbitMQ uri.

    Parameters
    ----------
    rabbit_uri: str
        URI of RabbitMQ service ('amqp://$user:$pass@$host:$port/$vhost')

    Returns
    -------
    dict[str]
        The vhost defined in the RabbitMQ URI
    """
    (user_details, location_details) = rabbit_uri.split("@", 1)
    (user, password) = user_details.split("/")[-1].split(":", 1)
    (host, remainder) = location_details.split(":", 1)
    port, vhost = remainder.split("/", 1)
    return {
        "user": user,
        "password": password,
        "host": host,
        "port": port,
        "vhost": vhost,
    }



import click
from colorama import Fore, Style

from vantage6.common import info, error, check_config_writeable
from vantage6.common.globals import InstanceType
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.cli.configuration_wizard import configuration_wizard
from vantage6.cli.utils import check_config_name_allowed, prompt_config_name


@click.command()
@click.option(
    "-n", "--name", default=None, help="name of the configuration you want to use."
)
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user", "system_folders", flag_value=False, default=DEFAULT_SERVER_SYSTEM_FOLDERS
)
def cli_algo_store_new(name: str, system_folders: bool) -> None:
    """
    Create a new server configuration.
    """
    name = prompt_config_name(name)

    # check if name is allowed for docker volume, else exit
    check_config_name_allowed(name)

    # check that this config does not exist
    try:
        if AlgorithmStoreContext.config_exists(name, system_folders):
            error(f"Configuration {Fore.RED}{name}{Style.RESET_ALL} already " "exists!")
            exit(1)
    except Exception as e:
        error(e)
        exit(1)

    # Check that we can write in this folder
    if not check_config_writeable(system_folders):
        error("Your user does not have write access to all folders. Exiting")
        info(
            f"Create a new server using '{Fore.GREEN}v6 algorithm-store new "
            f"--user{Style.RESET_ALL}' instead!"
        )
        exit(1)

    # create config in ctx location
    cfg_file = configuration_wizard(InstanceType.ALGORITHM_STORE, name, system_folders)
    info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")

    flag = "" if system_folders else "--user"
    info(
        f"You can start the algorithm store by running {Fore.GREEN}v6 "
        f"algorithm-store start {flag}{Style.RESET_ALL}"
    )



from docker.client import DockerClient
from colorama import Fore, Style

from vantage6.common import error, info
from vantage6.common.globals import APPNAME
from vantage6.common.docker.addons import remove_container, get_container
from vantage6.common.context import AppContext

from vantage6.cli.context.server import ServerContext


def get_server_context(
    name: str, system_folders: bool, ctx_class: AppContext
) -> AppContext:
    """
    Load the server context from the configuration file.

    Parameters
    ----------
    name : str
        Name of the server to inspect
    system_folders : bool
        Wether to use system folders or if False, the user folders
    ctx_class : AppContext
        Context class to be used. Derivative of AppContext class

    Returns
    -------
    ServerContext
        Server context object
    """
    if not ctx_class.config_exists(name, system_folders):
        scope = "system" if system_folders else "user"
        error(
            f"Configuration {Fore.RED}{name}{Style.RESET_ALL} does not "
            f"exist in the {Fore.RED}{scope}{Style.RESET_ALL} folders!"
        )
        exit(1)

    # We do not want to log this here, we do this in the container and not on
    # the host. We only want CLI logging here.
    ctx_class.LOGGING_ENABLED = False

    # create server context, and initialize db
    ctx = ctx_class(name, system_folders=system_folders)

    return ctx


def stop_ui(client: DockerClient, ctx: ServerContext) -> None:
    """
    Check if the UI container is running, and if so, stop and remove it.

    Parameters
    ----------
    client : DockerClient
        Docker client
    ctx : ServerContext
        Server context object
    """
    ui_container_name = f"{APPNAME}-{ctx.name}-{ctx.scope}-ui"
    ui_container = get_container(client, name=ui_container_name)
    if ui_container:
        remove_container(ui_container, kill=True)
        info(
            f"Stopped the {Fore.GREEN}{ui_container_name}"
            f"{Style.RESET_ALL} User Interface container."
        )



import click

from vantage6.common import info
from vantage6.common.globals import APPNAME, DEFAULT_ALGO_STORE_IMAGE, InstanceType
from vantage6.common.docker.addons import pull_image
from vantage6.cli.common.start import (
    attach_logs,
    check_for_start,
    get_image,
    mount_config_file,
    mount_database,
    mount_source,
)
from vantage6.cli.globals import AlgoStoreGlobals
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.cli.common.decorator import click_insert_context


@click.command()
@click.option("--ip", default=None, help="IP address to listen on")
@click.option("-p", "--port", default=None, type=int, help="Port to listen on")
@click.option("-i", "--image", default=None, help="Algorithm store Docker image to use")
@click.option(
    "--keep/--auto-remove",
    default=False,
    help="Keep image after algorithm store has been stopped. Useful " "for debugging",
)
@click.option(
    "--mount-src",
    default="",
    help="Override vantage6 source code in container with the source"
    " code in this path",
)
@click.option(
    "--attach/--detach",
    default=False,
    help="Print server logs to the console after start",
)
@click_insert_context(InstanceType.ALGORITHM_STORE)
def cli_algo_store_start(
    ctx: AlgorithmStoreContext,
    ip: str,
    port: int,
    image: str,
    keep: bool,
    mount_src: str,
    attach: bool,
) -> None:
    """
    Start the algorithm store server.
    """
    info("Starting algorithm store...")
    docker_client = check_for_start(ctx, InstanceType.ALGORITHM_STORE)

    image = get_image(image, ctx, "algorithm-store", DEFAULT_ALGO_STORE_IMAGE)

    info("Pulling algorithm store image...")
    pull_image(docker_client, image)

    config_file = "/mnt/config.yaml"
    mounts = mount_config_file(ctx, config_file)

    src_mount = mount_source(mount_src)
    if src_mount:
        mounts.append(src_mount)

    mount, environment_vars = mount_database(ctx, InstanceType.ALGORITHM_STORE)
    if mount:
        mounts.append(mount)

    # The `ip` and `port` refer here to the ip and port within the container.
    # So we do not really care that is it listening on all interfaces.
    internal_port = 5000
    cmd = (
        f"uwsgi --http :{internal_port} --gevent 1000 --http-websockets "
        "--master --callable app --disable-logging "
        "--wsgi-file /vantage6/vantage6-algorithm-store/vantage6/algorithm"
        f"/store/wsgi.py --pyargv {config_file}"
    )
    info(cmd)

    info("Run Docker container")
    port_ = str(port or ctx.config["port"] or AlgoStoreGlobals.PORT)
    container = docker_client.containers.run(
        image,
        command=cmd,
        mounts=mounts,
        detach=True,
        labels={
            f"{APPNAME}-type": InstanceType.ALGORITHM_STORE,
            "name": ctx.config_file_name,
        },
        environment=environment_vars,
        ports={f"{internal_port}/tcp": (ip, port_)},
        name=ctx.docker_container_name,
        auto_remove=not keep,
        tty=True,
    )

    info(f"Success! container id = {container.id}")

    if attach:
        attach_logs(container, InstanceType.ALGORITHM_STORE)



import questionary as q
from colorama import Fore, Style
import click
from typing import Iterable
import docker


from vantage6.common import warning, error
from vantage6.common.globals import APPNAME, InstanceType, STRING_ENCODING
from vantage6.cli.context import select_context_class


def get_server_name(
    name: str,
    system_folders: bool,
    running_server_names: list[str],
    instance_type: InstanceType,
) -> str:
    """
    Get the version of a running server.

    Parameters
    ----------
    name : str
        Name of the server to get the version from
    system_folders : bool
        Whether to use system folders or not
    running_server_names : list[str]
        The names of the running servers
    instance_type : InstanceType
        The type of instance to get the running servers from
    """

    if not name:
        if not running_server_names:
            error(
                f"No {instance_type}s are running! You can only check the version for "
                f"{instance_type}s that are running"
            )
            exit(1)
        name = q.select(
            f"Select the {instance_type} you wish to inspect:",
            choices=running_server_names,
        ).ask()
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}"
    return name


def get_running_servers(
    client: docker.DockerClient, instance_type: InstanceType
) -> list[str]:
    """Get the running servers of a certain type.

    Parameters
    ----------
    client : docker.DockerClient
        The docker client to use
    instance_type : InstanceType
        The type of instance to get the running servers from

    Returns
    -------
    list[str]
        The names of the running servers
    """
    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={instance_type}"}
    )
    return [server.name for server in running_servers]


def get_server_configuration_list(instance_type: InstanceType.SERVER) -> None:
    """
    Print list of available server configurations.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to get the configurations for
    """
    client = docker.from_env()
    ctx_class = select_context_class(instance_type)

    running_server_names = get_running_servers(client, instance_type)
    header = "\nName" + (21 * " ") + "Status" + (10 * " ") + "System/User"

    click.echo(header)
    click.echo("-" * len(header))

    running = Fore.GREEN + "Running" + Style.RESET_ALL
    stopped = Fore.RED + "Not running" + Style.RESET_ALL

    # system folders
    configs, f1 = ctx_class.available_configurations(system_folders=True)
    for config in configs:
        status = (
            running
            if f"{APPNAME}-{config.name}-system-{instance_type}" in running_server_names
            else stopped
        )
        click.echo(f"{config.name:25}" f"{status:25} System ")

    # user folders
    configs, f2 = ctx_class.available_configurations(system_folders=False)
    for config in configs:
        status = (
            running
            if f"{APPNAME}-{config.name}-user-{instance_type}" in running_server_names
            else stopped
        )
        click.echo(f"{config.name:25}" f"{status:25} User   ")

    click.echo("-" * 85)
    if len(f1) + len(f2):
        warning(f"{Fore.RED}Failed imports: {len(f1)+len(f2)}{Style.RESET_ALL}")


def print_log_worker(logs_stream: Iterable[bytes]) -> None:
    """
    Print the logs from the logs stream.

    Parameters
    ----------
    logs_stream : Iterable[bytes]
        Output of the container.attach() method
    """
    for log in logs_stream:
        try:
            print(log.decode(STRING_ENCODING), end="")
        except UnicodeDecodeError:
            print(
                "ERROR DECODING LOGS!!! Printing raw bytes. Please check the logs in "
                "the container."
            )
            print(log)



import click
import questionary as q

from vantage6.common import error, info, check_config_writeable
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.configuration_wizard import NodeConfigurationManager
from vantage6.cli.node.common import select_node


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--api-key", default=None, help="New API key")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders rather than " "user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in user folders rather than "
    "system folders. This is the default",
)
def cli_node_set_api_key(name: str, api_key: str, system_folders: bool) -> None:
    """
    Put a new API key into the node configuration file
    """
    # select node name
    name = select_node(name, system_folders)

    # Check that we can write in the config folder
    if not check_config_writeable(system_folders):
        error("Your user does not have write access to all folders. Exiting")
        exit(1)

    if not api_key:
        api_key = q.text("Please enter your new API key:").ask()

    # get configuration manager
    ctx = NodeContext(name, system_folders=system_folders)
    conf_mgr = NodeConfigurationManager.from_file(ctx.config_file)

    # set new api key, and save the file
    ctx.config["api_key"] = api_key
    conf_mgr.put(ctx.config)
    conf_mgr.save(ctx.config_file)
    info("Your new API key has been uploaded to the config file " f"{ctx.config_file}.")



import click
from colorama import Fore, Style

from vantage6.common import error, info, check_config_writeable
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.configuration_wizard import configuration_wizard
from vantage6.cli.utils import check_config_name_allowed, prompt_config_name
from vantage6.common.globals import InstanceType


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Store this configuration in the system folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Store this configuration in the user folders. This is the " "default",
)
def cli_node_new_configuration(name: str, system_folders: bool) -> None:
    """
    Create a new node configuration.

    Checks if the configuration already exists. If this is not the case
    a questionnaire is invoked to create a new configuration file.
    """
    name = prompt_config_name(name)
    # check if config name is allowed docker name
    check_config_name_allowed(name)

    # check that this config does not exist
    if NodeContext.config_exists(name, system_folders):
        error(f"Configuration {name} already exists!")
        exit(1)

    # Check that we can write in this folder
    if not check_config_writeable(system_folders):
        error("Cannot write configuration file. Exiting...")
        exit(1)

    # create config in ctx location
    flag = "--system" if system_folders else ""
    cfg_file = configuration_wizard(InstanceType.NODE, name, system_folders)
    info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")
    info(
        f"You can start the node by running "
        f"{Fore.GREEN}v6 node start {flag}{Style.RESET_ALL}"
    )



from pathlib import Path
import click

from vantage6.cli.utils import info
from vantage6.cli.dev.create import create_demo_network
from vantage6.cli.dev.start import start_demo_network
from vantage6.cli.dev.stop import stop_demo_network
from vantage6.cli.dev.remove import remove_demo_network
from vantage6.cli.utils import prompt_config_name, check_config_name_allowed
from vantage6.cli.test.feature_tester import cli_test_features


@click.command()
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your development setup"
)
@click.option(
    "--server-url",
    type=str,
    default="http://host.docker.internal",
    help="Server URL to point to. If you are using Docker Desktop, "
    "the default http://host.docker.internal should not be changed.",
)
@click.option(
    "-i", "--image", type=str, default=None, help="Server Docker image to use"
)
@click.option(
    "--keep",
    type=bool,
    default=False,
    help="Keep the dev network after finishing the test",
)
@click.option(
    "--extra-server-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional server "
    "configuration. This will be appended to the server "
    "configuration file",
)
@click.option(
    "--extra-node-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional node configuration. This will be"
    " appended to each of the node configuration files",
)
@click.pass_context
def cli_test_integration(
    click_ctx: click.Context,
    name: str,
    server_url: str,
    image: str,
    keep: bool = False,
    extra_server_config: Path = None,
    extra_node_config: Path = None,
) -> list[dict]:
    """
    Create dev network and run diagnostic checks on it.

    This is a full integration test of the vantage6 network. It will create
    a test server with some nodes using the `vdev` commands, and then run the
    v6-diagnostics algorithm to test all functionality.
    """
    # get name for the development setup - if not given - and check if it is
    # allowed
    name = prompt_config_name(name)
    check_config_name_allowed(name)

    # create server & node configurations and create test resources (
    # collaborations, organizations, etc)
    click_ctx.invoke(
        create_demo_network,
        name=name,
        num_nodes=3,
        server_url=server_url,
        server_port=5000,
        image=image,
        extra_server_config=extra_server_config,
        extra_node_config=extra_node_config,
    )

    # start the server and nodes
    click_ctx.invoke(
        start_demo_network,
        name=name,
        system_folders=True,
        server_image=image,
        node_image=image,
    )

    # run the diagnostic tests
    # TODO the username and password should be coordinated with the vdev
    # command - at present it spits out this username/password combination by
    # default but both should be defined in the same place
    # TODO VPN testing is always excluded - allow to include it with a flag
    # when vdev commands can handle extra config parameters
    diagnose_results = click_ctx.invoke(
        cli_test_features,
        host="http://localhost",
        port=5000,
        api_path="/api",
        username="org_1-admin",
        password="password",
        collaboration=1,
        organizations=None,
        all_nodes=True,
        online_only=False,
        no_vpn=True,
    )

    # clean up the test resources
    click_ctx.invoke(stop_demo_network, name=name, system_folders=True)
    if not keep:
        click_ctx.invoke(remove_demo_network, name=name, system_folders=True)
    else:
        info(
            "Keeping the demo network {name}. You can start it with `v6 dev "
            "start-demo-network`"
        )

    return diagnose_results



"""
Common functions that are used in node CLI commands
"""

import questionary as q
import docker
from colorama import Fore, Style

from vantage6.common import error, info, debug
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.client import UserClient

from vantage6.cli.context.node import NodeContext
from vantage6.cli.configuration_wizard import select_configuration_questionaire


def create_client(ctx: NodeContext) -> UserClient:
    """
    Create a client instance.

    Parameters
    ----------
    ctx : NodeContext
        Context of the node loaded from the configuration file
    Returns
    -------
    UserClient
        vantage6 client
    """
    host = ctx.config["server_url"]
    # if the server is run locally, we need to use localhost here instead of
    # the host address of docker
    if host in ["http://host.docker.internal", "http://172.17.0.1"]:
        host = "http://localhost"
    port = ctx.config["port"]
    api_path = ctx.config["api_path"]
    info(f"Connecting to server at '{host}:{port}{api_path}'")
    return UserClient(host, port, api_path, log_level="warn")


def create_client_and_authenticate(ctx: NodeContext) -> UserClient:
    """
    Generate a client and authenticate with the server.

    Parameters
    ----------
    ctx : NodeContext
        Context of the node loaded from the configuration file

    Returns
    -------
    UserClient
        vantage6 client
    """
    client = create_client(ctx)

    username = q.text("Username:").ask()
    password = q.password("Password:").ask()

    try:
        client.authenticate(username, password)

    except Exception as exc:
        error("Could not authenticate with server!")
        debug(exc)
        exit(1)

    return client


def select_node(name: str, system_folders: bool) -> tuple[str, str]:
    """
    Let user select node through questionnaire if name is not given.

    Returns
    -------
    str
        Name of the configuration file
    """
    name = (
        name
        if name
        else select_configuration_questionaire(InstanceType.NODE, system_folders)
    )

    # raise error if config could not be found
    if not NodeContext.config_exists(name, system_folders):
        error(
            f"The configuration {Fore.RED}{name}{Style.RESET_ALL} could "
            f"not be found."
        )
        exit(1)
    return name


def find_running_node_names(client: docker.DockerClient) -> list[str]:
    """
    Returns a list of names of running nodes.

    Parameters
    ----------
    client : docker.DockerClient
        Docker client instance

    Returns
    -------
    list[str]
        List of names of running nodes
    """
    running_nodes = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.NODE}"}
    )
    return [node.name for node in running_nodes]



import click
import docker

from vantage6.common import error
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import InstanceType
from vantage6.cli.common.utils import get_server_name, get_running_servers
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli import __version__


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user", "system_folders", flag_value=False, default=DEFAULT_SERVER_SYSTEM_FOLDERS
)
def cli_server_version(name: str, system_folders: bool) -> None:
    """
    Print the version of the vantage6 server.
    """
    check_docker_running()
    client = docker.from_env()

    running_server_names = get_running_servers(client, InstanceType.SERVER)

    name = get_server_name(
        name, system_folders, running_server_names, InstanceType.SERVER
    )

    if name in running_server_names:
        container = client.containers.get(name)
        version = container.exec_run(cmd="vserver-local version", stdout=True)
        click.echo({"server": version.output.decode("utf-8"), "cli": __version__})
    else:
        error(f"Server {name} is not running! Cannot provide version...")



import unittest

from pathlib import Path
from unittest.mock import patch, MagicMock

from vantage6.cli.configuration_wizard import (
    node_configuration_questionaire,
    server_configuration_questionaire,
    configuration_wizard,
    select_configuration_questionaire,
)
from vantage6.common.globals import InstanceType

module_path = "vantage6.cli.configuration_wizard"


class WizardTest(unittest.TestCase):
    @staticmethod
    def prompts(*args, **kwargs):
        result = {}
        for arg in args[0]:
            name = arg["name"]
            if name == "default":  # default db path
                result[name] = "/some/path/db.sqlite"
            else:
                if "default" in arg:
                    result[name] = arg["default"]
                else:
                    result[name] = None
        return result

    def test_node_wizard(self):
        """An error is printed when docker is not running"""

        with patch(f"{module_path}.q") as q:
            q.prompt.side_effect = self.prompts
            q.confirm.return_value.ask.side_effect = [
                True,  # add a database
                False,  # don't enable two-factor authentication
                True,  # add VPN server
                True,  # add algorithm policies
                "some-image",  # algorithm image to whitelist
                False,  # don't add another algorithm image
                False,  # don't abort if no server connection is made to pull
                # collaboration settings
                True,  # Enable encryption
            ]
            dirs = MagicMock(data="/")
            config = node_configuration_questionaire(dirs, "iknl")

        keys = [
            "api_key",
            "server_url",
            "port",
            "api_path",
            "task_dir",
            "databases",
            "logging",
            "encryption",
            "vpn_subnet",
        ]
        for key in keys:
            self.assertIn(key, config)
        nested_keys = [["policies", "allowed_algorithms"]]
        for nesting in nested_keys:
            current_config = config
            for key in nesting:
                self.assertIn(key, current_config)
                current_config = current_config[key]

    def test_server_wizard(self):
        with patch(f"{module_path}.q") as q:
            q.prompt.side_effect = self.prompts
            q.confirm.return_value.ask.side_effect = [
                True,
                True,
                True,
                True,
                True,
                True,
                False,
            ]

            config = server_configuration_questionaire("vantage6")

            keys = [
                "description",
                "ip",
                "port",
                "api_path",
                "uri",
                "allow_drop_all",
                "jwt_secret_key",
                "logging",
                "vpn_server",
                "rabbitmq",
                "two_factor_auth",
                "algorithm_stores",
            ]

            for key in keys:
                self.assertIn(key, config)

    @patch(f"{module_path}.node_configuration_questionaire")
    @patch(f"{module_path}.server_configuration_questionaire")
    @patch(f"{module_path}.ServerConfigurationManager")
    @patch(f"{module_path}.NodeConfigurationManager")
    @patch("vantage6.cli.configuration_wizard.AppContext")
    def test_configuration_wizard_interface(
        self, context, node_m, server_m, server_q, node_q
    ):
        context.instance_folders.return_value = {"config": "/some/path/"}

        file_ = configuration_wizard(InstanceType.NODE, "vtg6", False)
        self.assertEqual(Path("/some/path/vtg6.yaml"), file_)

        file_ = configuration_wizard(InstanceType.SERVER, "vtg6", True)
        self.assertEqual(Path("/some/path/vtg6.yaml"), file_)

    @patch("vantage6.cli.configuration_wizard.AppContext.available_configurations")
    def test_select_configuration(self, available_configurations):
        config = MagicMock()
        config.name = "vtg6"

        available_configurations.return_value = [[config], []]

        with patch(f"{module_path}.q") as q:
            q.select.return_value.ask.return_value = "vtg6"
            name = select_configuration_questionaire(InstanceType.NODE, True)

        self.assertEqual(name, "vtg6")



import os
from pathlib import Path

import click
import questionary as q
from copier import run_copy

from vantage6.cli.globals import ALGORITHM_TEMPLATE_REPO
from vantage6.cli.utils import info


@click.command()
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your new algorithm"
)
@click.option(
    "-d",
    "--dir",
    "directory",
    default=None,
    type=str,
    help="Directory to put the algorithm into",
)
def cli_algorithm_create(name: str, directory: str) -> dict:
    """Creates a personalized template for a new algorithm

    By answering a number of questions, a template will be created that will
    simplify the creation of a new algorithm. The goal is that the algorithm
    developer only focuses on the algorithm code rather than fitting it to
    the vantage6 infrastructure.

    The created template will contain a Python package with a Dockerfile that
    can be used to build an appropriate Docker image that can be used as a
    vantage6 algorithm.
    """
    if not name:
        name = q.text("Name of your new algorithm:").ask()

    if not directory:
        default_dir = str(Path(os.getcwd()) / name)
        directory = q.text(
            "Directory to put the algorithm in:", default=default_dir
        ).ask()

    # Create the template. The `unsafe` flag is used to allow running a Python script
    # after creating the template that cleans up some things.
    run_copy(
        ALGORITHM_TEMPLATE_REPO, directory, data={"algorithm_name": name}, unsafe=True
    )
    info("Template created!")
    info(f"You can find your new algorithm in: {directory}")



import subprocess
import click

from vantage6.cli.context.server import ServerContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.server.start import cli_server_start


@click.command()
@click_insert_context(type_="server")
@click.option(
    "--server-image", type=str, default=None, help="Server Docker image to use"
)
@click.option("--node-image", type=str, default=None, help="Node Docker image to use")
@click.pass_context
def start_demo_network(
    click_ctx: click.Context, ctx: ServerContext, server_image: str, node_image: str
) -> None:
    """Starts running a demo-network.

    Select a server configuration to run its demo network. You should choose a
    server configuration that you created earlier for a demo network. If you
    have not created a demo network, you can run `vdev create-demo-network` to
    create one.
    """
    # run the server
    click_ctx.invoke(
        cli_server_start,
        ctx=ctx,
        ip=None,
        port=None,
        image=server_image,
        start_ui=False,
        ui_port=None,
        start_rabbitmq=False,
        rabbitmq_image=None,
        keep=True,
        mount_src="",
        attach=False,
    )

    # run all nodes that belong to this server
    configs, _ = NodeContext.available_configurations(system_folders=False)
    node_names = [
        config.name for config in configs if f"{ctx.name}_node_" in config.name
    ]
    for name in node_names:
        cmd = ["v6", "node", "start", "--name", name]
        if node_image:
            cmd.extend(["--image", node_image])
        subprocess.run(cmd)



import os
from threading import Thread
import time
import docker
import enum
from docker.client import DockerClient
from docker.models.containers import Container
from colorama import Fore, Style
from sqlalchemy.engine.url import make_url

from vantage6.common import error, info, warning
from vantage6.common.context import AppContext
from vantage6.common.globals import InstanceType, APPNAME, DEFAULT_DOCKER_REGISTRY
from vantage6.common.docker.addons import check_docker_running
from vantage6.cli.context import AlgorithmStoreContext, ServerContext
from vantage6.cli.common.utils import print_log_worker
from vantage6.cli.utils import check_config_name_allowed
from vantage6.cli.globals import ServerGlobals, AlgoStoreGlobals


def check_for_start(ctx: AppContext, type_: InstanceType) -> DockerClient:
    """
    Check if all requirements are met to start the instance.

    Parameters
    ----------
    ctx : AppContext
        The context object
    type_ : InstanceType
        The type of instance to check for

    Returns
    -------
    DockerClient
        A Docker client instance
    """
    # will print an error if not
    check_docker_running()

    info("Finding Docker daemon.")
    docker_client = docker.from_env()

    # check if name is allowed for docker volume, else exit
    check_config_name_allowed(ctx.name)

    # check that this server is not already running
    running_servers = docker_client.containers.list(
        filters={"label": f"{APPNAME}-type={type_}"}
    )
    for server in running_servers:
        if server.name == f"{APPNAME}-{ctx.name}-{ctx.scope}-{type_}":
            error(f"Server {Fore.RED}{ctx.name}{Style.RESET_ALL} " "is already running")
            exit(1)
    return docker_client


def get_image(
    image: str, ctx: AppContext, custom_image_key: str, default_image: str
) -> str:
    """
    Get the image name for the given instance type.

    Parameters
    ----------
    image : str | None
        The image name to use if specified
    ctx : AppContext
        The context object
    custom_image_key : str
        The key to look for in the config file
    default_image : str
        The default image name

    Returns
    -------
    str
        The image name to use
    """
    # Determine image-name. First we check if the option --image has been used.
    # Then we check if the image has been specified in the config file, and
    # finally we use the default settings from the package.
    if image is None:
        custom_images: dict = ctx.config.get("images")
        if custom_images:
            image = custom_images.get(custom_image_key)
        if not image:
            image = f"{DEFAULT_DOCKER_REGISTRY}/{default_image}"
    return image


def mount_config_file(ctx: AppContext, config_file: str) -> list[docker.types.Mount]:
    """
    Mount the config file in the container.

    Parameters
    ----------
    ctx : AppContext
        The context object
    config_file : str
        The path to the config file

    Returns
    -------
    list[docker.types.Mount]
        The mounts to use
    """
    info("Creating mounts")
    return [docker.types.Mount(config_file, str(ctx.config_file), type="bind")]


def mount_source(mount_src: str) -> docker.types.Mount:
    """
    Mount the vantage6 source code in the container.

    Parameters
    ----------
    mount_src : str
        The path to the source code

    Returns
    -------
    docker.types.Mount | None
        The mount to use
    """
    if mount_src:
        mount_src = os.path.abspath(mount_src)
        return docker.types.Mount("/vantage6", mount_src, type="bind")


def mount_database(
    ctx: ServerContext | AlgorithmStoreContext, type_: InstanceType
) -> tuple[docker.types.Mount, dict]:
    """
    Mount database in the container if it is file-based (e.g. a SQLite DB).

    Parameters
    ----------
    ctx : AppContext
        The context object
    type_ : InstanceType
        The type of instance to mount the database for

    Returns
    -------
    docker.types.Mount | None
        The mount to use
    dict | None
        The environment variables to use
    """
    # FIXME: code duplication with cli_server_import()
    # try to mount database
    uri = ctx.config["uri"]
    url = make_url(uri)
    environment_vars = None
    mount = None

    # If host is None, we're dealing with a file-based DB, like SQLite
    if url.host is None:
        db_path = url.database

        if not os.path.isabs(db_path):
            # We're dealing with a relative path here -> make it absolute
            db_path = ctx.data_dir / url.database

        basename = os.path.basename(db_path)
        dirname = os.path.dirname(db_path)
        os.makedirs(dirname, exist_ok=True)

        # we're mounting the entire folder that contains the database
        mount = docker.types.Mount("/mnt/database/", dirname, type="bind")

        if type_ == InstanceType.SERVER:
            environment_vars = {
                ServerGlobals.DB_URI_ENV_VAR.value: f"sqlite:////mnt/database/{basename}",
                ServerGlobals.CONFIG_NAME_ENV_VAR.value: ctx.config_file_name,
            }
        elif type_ == InstanceType.ALGORITHM_STORE:
            environment_vars = {
                AlgoStoreGlobals.DB_URI_ENV_VAR.value: f"sqlite:////mnt/database/{basename}",
                AlgoStoreGlobals.CONFIG_NAME_ENV_VAR.value: ctx.config_file_name,
            }
    else:
        warning(
            f"Database could not be transferred, make sure {url.host} "
            "is reachable from the Docker container"
        )
        info("Consider using the docker-compose method to start a server")

    return mount, environment_vars


def attach_logs(container: Container, type_: InstanceType) -> None:
    """
    Attach container logs to the console if specified.

    Parameters
    ----------
    container : Container
        The container to attach the logs from
    type_ : InstanceType
        The type of instance to attach the logs for
    """
    if isinstance(type_, enum.Enum):
        type_ = type_.value
    logs = container.attach(stream=True, logs=True, stdout=True)
    Thread(target=print_log_worker, args=(logs,), daemon=True).start()
    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            info("Closing log file. Keyboard Interrupt.")
            info(
                "Note that your server is still running! Shut it down "
                f"with {Fore.RED}v6 {type_} stop{Style.RESET_ALL}"
            )
            exit(0)
